from __future__ import annotations

import urllib
from typing import TYPE_CHECKING
from typing import Any

import requests as r
from pyjstat import pyjstat

if TYPE_CHECKING:
    import pandas as pd

    from statbank.api_types import QueryPartType
    from statbank.api_types import QueryWholeType


from statbank.statbank_logger import logger

# Getting data from Statbank

STATBANK_TABLE_ID_LENGTH = 5
REQUESTS_OK_RETURN = 200


def apidata(
    id_or_url: str = "",
    payload: QueryWholeType | None = None,
    include_id: bool = False,
) -> pd.DataFrame:
    """Get the contents of a published statbank-table as a pandas Dataframe, specifying a query to limit the return.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
        payload (QueryWholeType | None): a dict in the shape of a QueryWhole, to include with the request, can be copied from the statbank-webpage.
        include_id (bool): If you want to include "codes" in the dataframe, set this to True

    Returns:
        pd.DataFrame: The table-content

    Raises:
        ValueError: If the first parameter is not recognized as a statbank ID or a direct url.
    """
    if payload is None:
        payload_now: QueryWholeType = {
            "query": [],
            "response": {"format": "json-stat2"},
        }
    else:
        payload_now = payload
    if len(id_or_url) == STATBANK_TABLE_ID_LENGTH and id_or_url.isdigit():
        url = f"https://data.ssb.no/api/v0/no/table/{id_or_url}/"
    else:
        test_url = urllib.parse.urlparse(id_or_url)
        if not all([test_url.scheme, test_url.netloc]):
            error_msg = (
                "First parameter not recognized as a statbank ID or a direct url"
            )
            raise ValueError(error_msg)
        url = id_or_url

    logger.info(url)
    # Spør APIet om å få resultatet med requests-biblioteket
    resultat = r.post(url, json=payload_now, timeout=10)
    resultat.raise_for_status()
    # Putt teksten i resultatet inn i ett pyjstat-datasett-objekt
    dataset_pyjstat = pyjstat.Dataset.read(resultat.text)
    # Skriv pyjstat-objektet ut som en pandas dataframe
    table_data: pd.DataFrame = dataset_pyjstat.write("dataframe")
    # Om man ønsker IDen påført dataframen, så er vi fancy
    if include_id:
        table_data_ids = dataset_pyjstat.write("dataframe", naming="id")
        skip = 0
        for i, col in enumerate(table_data_ids.columns):
            insert_at = (i + 1) * 2 - 1 - skip
            df_col_tocompare = table_data.iloc[:, insert_at - 1]
            # Sett inn kolonne på rett sted, avhengig av at navnet ikke er brukt
            # og at nabokolonnen ikke har samme verdier.
            if col not in table_data.columns and not table_data_ids[col].equals(
                df_col_tocompare,
            ):
                table_data.insert(insert_at, col, table_data_ids[col])
            # Indexen må justeres, om vi lar være å skrive inn kolonnen
            else:
                skip += 1
    return table_data.convert_dtypes()


def apidata_all(id_or_url: str = "", include_id: bool = False) -> pd.DataFrame:
    """Get ALL the contents of a published statbank-table as a pandas Dataframe.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
        include_id (bool): If you want to include "codes" in the dataframe, set this to True

    Returns:
        pd.DataFrame: Table-content
    """
    return apidata(id_or_url, apidata_query_all(id_or_url), include_id=include_id)


def apimetadata(id_or_url: str = "") -> dict[str, Any]:
    """Get the metadata of a published statbank-table as a dict.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".

    Returns:
        dict[str, Any]: The metadata of the table as the json returned from the API-get-request.

    Raises:
        ValueError: If the first parameter is not recognized as a statbank ID or a direct url.
    """
    if len(id_or_url) == STATBANK_TABLE_ID_LENGTH and id_or_url.isdigit():
        url = f"https://data.ssb.no/api/v0/no/table/{id_or_url}/"
    else:
        test_url = urllib.parse.urlparse(id_or_url)
        if not all([test_url.scheme, test_url.netloc]):
            error_msg = (
                "First parameter not recognized as a statbank ID or a direct url"
            )
            raise ValueError(error_msg)
        url = id_or_url
    res = r.get(url, timeout=5)
    res.raise_for_status()
    meta: dict[str, Any] = res.json()
    return meta


def apicodelist(
    id_or_url: str = "",
    codelist_name: str = "",
) -> dict[str, str] | dict[str, dict[str, str]]:
    """Get one specific or all the codelists of a published statbank-table as a dict or nested dicts.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
        codelist_name (str): The name of the specific codelist to get.

    Returns:
        dict[str, str] | dict[str, dict[str, str]]: The codelist of the table as a dict or a nested dict.

    Raises:
        ValueError: If the specified codelist_name is not in the returned metadata.
    """
    metadata = apimetadata(id_or_url)
    results = {}
    for col in metadata["variables"]:
        results[col["code"]] = dict(zip(col["values"], col["valueTexts"]))
    if codelist_name == "":
        return results
    if codelist_name in results:
        return results[codelist_name]
    for col in metadata["variables"]:
        if codelist_name == col["text"]:
            return dict(zip(col["values"], col["valueTexts"]))
    col_names = ", ".join([col["code"] for col in metadata["variables"]])
    error_msg = f"Cant find {codelist_name} among the available names: {col_names}"
    raise ValueError(error_msg)


def apidata_query_all(id_or_url: str = "") -> QueryWholeType:
    """Builds a query for ALL THE DATA in a table based on a request for metadata on the table.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".

    Returns:
        QueryWholeType: The prepared query based on all the codes in the table.
    """
    meta = apimetadata(id_or_url)["variables"]
    code_list: list[QueryPartType] = []
    for code in meta:
        tmp: QueryPartType = {"code": "", "selection": {"filter": "", "values": []}}
        for k, v in code.items():
            if k == "code":
                tmp["code"] = v
            if k == "values":
                tmp["selection"] = {"filter": "item", "values": v}
        code_list += [tmp]
    return {"query": code_list, "response": {"format": "json-stat2"}}


# Credit: https://github.com/sehyoun/SSB_API_helper/blob/master/src/ssb_api_helper.py
def apidata_rotate(
    df: pd.DataFrame,
    ind: str = "year",
    val: str = "value",
) -> pd.DataFrame:
    """Rotate the dataframe so that years are used as the index.

    Args:
        df (pd.DataFrame): dataframe (from <get_from_ssb> function)
        ind (str): string of column name denoting time
        val (str): string of column name denoting values

    Returns:
        pd.DataFrame: pivoted dataframe
    """
    return df.pivot_table(
        index=ind,
        values=val,
        columns=[i for i in df.columns if i not in (ind, val)],
    )



import base64
import getpass
import json
import os

import requests as r
from dapla import AuthClient


class StatbankAuth:
    """Parent class for shared behavior between Statbankens "Transfer-API" and "Uttaksbeskrivelse-API".

    Methods:
        _build_headers() -> dict:
            Creates dict of headers needed in request to talk to Statbank-API
        _build_auth() -> str:
            Gets key from environment and encrypts password with key, combines it with username into expected Authentication header.
        _encrypt_request() -> str:
            Encrypts password with key from local service, url for service should be environment variables. Password is not possible to send into function. Because safety.
        _build_urls() -> dict:
            Urls will differ based environment variables, returns a dict of urls.
        __init__():

            is not implemented, as Transfer and UttrekksBeskrivelse both add their own.
    """

    def __init__(self) -> None:
        """This init will never be used directly, as this class is always inherited from.

        This is for typing with Mypy.
        """

    def _build_headers(self) -> dict[str, str]:
        return {
            "Authorization": self._build_auth(),
            "Content-Type": "multipart/form-data; boundary=12345",
            "Connection": "keep-alive",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept": r"*/*",
            "User-Agent": self._build_user_agent(),
        }

    @staticmethod
    def check_env() -> str:
        """Check if you are on Dapla or in prodsone.

        Simplified terribly by the addition of env vars for this, keeping this method for legacy reasons.

        Returns:
            str: "DAPLA" if on dapla, "PROD" if you are in prodsone.
        """
        return os.environ.get("DAPLA_ENVIRONMENT", "TEST")

    @staticmethod
    def check_database() -> str:
        """Checks if we are in prod environment. And which statbank-database we are sending to."""
        db = "TEST"
        if os.environ.get("DAPLA_ENVIRONMENT", "TEST") == "PROD":
            db = "PROD"
        return db

    def _build_user_agent(self) -> str:
        envir = os.environ.get("DAPLA_ENVIRONMENT", "TEST")
        service = os.environ.get("DAPLA_SERVICE", "JUPYTERLAB")
        region = os.environ.get("DAPLA_REGION", "ON_PREM")

        user_agent = f"{envir}-{region}-{service}-"

        return user_agent + r.utils.default_headers()["User-agent"]

    def _build_auth(self) -> str:
        username_encryptedpassword = (
            bytes(
                self._get_user(),
                "UTF-8",
            )
            + bytes(":", "UTF-8")
            + bytes(json.loads(self._encrypt_request().text)["message"], "UTF-8")
        )
        return "Basic " + base64.b64encode(username_encryptedpassword).decode("utf8")

    @staticmethod
    def _get_user() -> str:
        return getpass.getpass("Lastebruker:")

    def _encrypt_request(self) -> r.Response:
        db = self.check_database()
        if AuthClient.is_ready():
            headers = {
                "Authorization": f"Bearer {AuthClient.fetch_personal_token()}",
                "Content-type": "application/json",
            }
        else:
            headers = {
                "Content-type": "application/json",
            }
        return r.post(
            os.environ.get("STATBANK_ENCRYPT_URL", "Cant find url in environ."),
            headers=headers,
            json={"message": getpass.getpass(f"Lastepassord ({db}):")},
            timeout=5,
        )

    @staticmethod
    def _build_urls() -> dict[str, str]:
        base_url = os.environ.get("STATBANK_BASE_URL", "Cant find url in environ.")
        end_urls = {
            "loader": "statbank/sos/v1/DataLoader?",
            "uttak": "statbank/sos/v1/uttaksbeskrivelse?",
            "gui": "lastelogg/gui/",
            "api": "lastelogg/api/",
        }
        return {k: base_url + v for k, v in end_urls.items()}



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

if TYPE_CHECKING:
    import pandas as pd

import datetime as dt
import json
import os
from pathlib import Path
from typing import TYPE_CHECKING

import ipywidgets as widgets
from IPython.display import display

if TYPE_CHECKING:
    from statbank.api_types import QueryWholeType
from statbank.apidata import apicodelist
from statbank.apidata import apidata
from statbank.apidata import apidata_all
from statbank.apidata import apidata_rotate
from statbank.apidata import apimetadata
from statbank.auth import StatbankAuth
from statbank.globals import APPROVE_DEFAULT_JIT
from statbank.globals import OSLO_TIMEZONE
from statbank.globals import STATBANK_TABLE_ID_LEN
from statbank.globals import TOMORROW
from statbank.globals import Approve
from statbank.globals import _approve_type_check
from statbank.statbank_logger import logger
from statbank.transfer import StatbankTransfer
from statbank.uttrekk import StatbankUttrekksBeskrivelse


class StatbankClient(StatbankAuth):
    """This is the main interface towards the rest of the statbank-package.

    An initialized client, an object of this class, will contain data/parameters
    that often is shared, among all transfers within a statistical production.
    Call methods under this client to:
    - transfer the data: .transfer()
    - only validate the data against a description: .validate()
    - get transfer/data description (filbeskrivelse): .get_description()
    - set the publish date with a datepicker: .date_picker() + .set_publish_date()
    - get published data from the external or internal API of statbanken: apidata_all() / apidata()

    Attributes:
        date (str): Date for publishing the transfer. Shape should be "yyyy-mm-dd",
            like "2022-01-01".
            Statbanken only allows publishing four months into the future?
        shortuser (str): The abbrivation of username at ssb. Three letters, like "cfc".
            If not specified,
            we will try to get this from daplas environement variables.
        cc (str): First person to be notified by email of transfer.
            Defaults to the same as "shortuser"
        bcc (str): Second person to be notified by email of transfer.
            Defaults to the same as "cc"
        overwrite (bool): False = no overwrite
            True = overwrite
        approve (Approve | str | int):
            0 = MANUAL approval
            1 = AUTOMATIC approval at transfer-time (immediately)
            2 = JIT (Just In Time), approval right before publishing time
        log (list[str]): Each "action" (method used) on the client is appended to the log.
            Nice to use for appending to your own logging after you are done,
            or printing it in a try-except-block to see what the last actions were,
            before error being raised.
    """

    def __init__(  # noqa: PLR0913
        self,
        date: str | dt.datetime = TOMORROW,
        shortuser: str = "",
        cc: str = "",
        bcc: str = "",
        overwrite: bool = True,
        approve: (
            int | str | Approve
        ) = APPROVE_DEFAULT_JIT,  # Changing back to 2, after wish from Rakel Gading
        check_username_password: bool = True,
    ) -> None:
        """Initialize the client, storing password etc. on the client."""
        self.shortuser = shortuser
        self.cc = cc
        self.bcc = bcc
        self.overwrite = overwrite
        self.approve = _approve_type_check(approve)
        self.check_username_password = check_username_password
        self._validate_params_init()
        self.__headers = self._build_headers()
        self.log: list[str] = []
        if isinstance(date, str):
            try:
                self.date: dt.datetime = dt.datetime.strptime(
                    date,
                    "%Y-%m-%d",
                ).astimezone(
                    OSLO_TIMEZONE,
                ) + dt.timedelta(
                    hours=1,
                )  # Compensate for setting the timezone, stop publishing date from moving
            except ValueError as e:
                error_msg = f"Loaduser parameter removed, please do not use it in your code. OR: {e}"
                raise ValueError(error_msg) from e
        else:
            self.date = date
        self._validate_date()
        self.date = self.date.replace(hour=8, minute=0, second=0, microsecond=0)
        if self.check_username_password:
            logger.info(
                "Checking filbeskrivelse of random tableid 05300 to double-check username & password early.",
            )
            self.get_description(
                "05300",
            )
        logger.info("Publishing date set to %s", self.date.isoformat("T", "seconds"))

    # Representation
    def __str__(self) -> str:
        """Print a human readable text of the clients attributes."""
        return f"""StatbankClient
        Publishing at {self.date}
        Shortuser {self.shortuser}
        Sending mail to {self.cc}
        And sending mail to {self.bcc}
        Overwrite set to {self.overwrite}
        Approve set to {self.approve}

        Log:
        """ + "\n\t".join(
            self.log,
        )

    def __repr__(self) -> str:
        """Represent the class with the necessary argument to replicate."""
        result = "StatbankClient("
        if self.date != TOMORROW:
            result += f'date = "{self.date.isoformat("T", "seconds")}", '
        if self.shortuser:
            result += f'shortuser = "{self.shortuser}", '
        if self.cc:
            result += f'cc = "{self.cc}", '
        if self.bcc:
            result += f', bcc = "{self.bcc}", '
        if not self.overwrite:
            result += f"overwrite = {self.overwrite}), "
        if self.approve != APPROVE_DEFAULT_JIT:
            result += f"approve = {self.approve}, "
        if self.check_username_password:
            result += f"check_username_password = {self.check_username_password}"
        result = result.strip(" ").strip(",")
        result += ")"
        return result

    # Publishing date handeling
    def date_picker(self) -> widgets.DatePicker:
        """Display a datapicker-widget.

        Assign it to a variable, that you after editing the date,
        pass into set_publish_date()
        date = client.datepicker()
        # Edit date
        client.set_publish_date(date)

        Returns:
            widgets.DatePicker: A datepicker widget from ipywidgets, with its date set to what the client currently holds.
        """
        datepicker = widgets.DatePicker(
            description="Publish-date",
            disabled=False,
            value=self.date,
        )
        display(datepicker)  # type: ignore[no-untyped-call]
        return datepicker

    def set_publish_date(self, date: dt.datetime | str | widgets.DatePicker) -> None:
        """Set the publishing date on the client.

        Takes the widget from date_picker assigned to a variable, which is probably the intended use.
        If sending a string, use the format 2000-12-31, you can also send in a datetime.
        Hours, minutes and seconds are replaced with statbankens publish time: 08:00:00

        Args:
            date (datetime): date-picker widget, or a date-string formatted as 2000-12-31

        Raises:
            TypeError: If the date-parameter is of type other than datetime, string, or ipywidgets.DatePicker.
        """
        if isinstance(date, widgets.DatePicker):
            date_date = dt.datetime.combine(
                date.value,
                dt.datetime.min.time(),
            ).astimezone(OSLO_TIMEZONE) + dt.timedelta(hours=1)
        elif isinstance(date, str):
            date_date = dt.datetime.strptime(date, "%Y-%m-%d").astimezone(
                OSLO_TIMEZONE,
            ) + dt.timedelta(hours=1)
        elif isinstance(date, dt.datetime):
            date_date = date
        else:
            error_msg = f"date-parameter is of type {type(date)} must be a string, datetime, or ipywidgets.DatePicker"
            raise TypeError(error_msg)

        self.date = date_date
        self.date = self.date.replace(hour=8, minute=0, second=0, microsecond=0)
        self._validate_date()
        logger.info("Publishing date set to: %s", self.date)
        self.log.append(
            f"Date set to {self.date.isoformat('T', 'seconds')} at {(dt.datetime.now().astimezone(OSLO_TIMEZONE) + dt.timedelta(hours=1)).isoformat('T', 'seconds')}",
        )

    # Descriptions
    def get_description(
        self,
        tableid: str = "00000",
    ) -> StatbankUttrekksBeskrivelse:
        """Get the "uttrekksbeskrivelse" for the tableid, which describes metadata.

        about shape of data to be transferred, and metadata about the table
        itself in Statbankens system, like ID, name and content of codelists.

        Args:
            tableid (str): The tableid of the "hovedtabell" in statbanken, a 5 digit string.

        Returns:
            StatbankUttrekksBeskrivelse: An instance of the class StatbankUttrekksBeskrivelse, which is comparable to the old "filbeskrivelse".
        """
        self._validate_params_action(tableid)
        self.log.append(
            f"Getting description for tableid {tableid} at {(dt.datetime.now().astimezone(OSLO_TIMEZONE,) + dt.timedelta(hours=1)).isoformat('T', 'seconds')}",
        )
        return StatbankUttrekksBeskrivelse(
            tableid=tableid,
            headers=self.__headers,
        )

    @staticmethod
    def read_description_json(json_path_or_str: str) -> StatbankUttrekksBeskrivelse:
        """Re-initializes a StatbankUttrekksBeskrivelse from a stored json file/string.

        Checks if provided string exists on disk, if it does, tries to load it as json.
        Otherwise expects you to provide a json-string that works for json.loads.
        Inserts first layer in json as attributes under a blank StatbankUttrekksBeskrivelse-object.

        Args:
            json_path_or_str (str): Either a path on local storage, or a loaded json-string

        Returns:
            StatbankUttrekksBeskrivelse: An instance of the class StatbankUttrekksBeskrivelse, which is comparable to the old "filbeskrivelse".
        """
        content = json_path_or_str
        try:
            try_path = json_path_or_str
            if Path(try_path).exists():
                with Path(try_path).open("r") as json_file:
                    content = json_file.read()
        except OSError as e:
            logger.debug(
                "Assuming you sent a json-string to open as description, cause that path does not exist. %s",
                str(e),
            )
        new = StatbankUttrekksBeskrivelse.__new__(StatbankUttrekksBeskrivelse)
        for k, v in json.loads(content).items():
            setattr(new, k, v)
        return new

    # Validation
    def validate(
        self,
        dfs: dict[str, pd.DataFrame],
        tableid: str = "00000",
        raise_errors: bool = False,
    ) -> dict[str, ValueError]:
        """Gets an "uttrekksbeskrivelse" and validates the data against this.

        All validation happens locally, so dont be afraid of any data
        being sent to statbanken using this method.

        Args:
            dfs (dict[str, pd.DataFrame): The data to validate in a dictionary of deltabell-names as keys and pandas-dataframes as values.
            tableid (str): The tableid of the "hovedtabell" in statbanken, a 5 digit string. Defaults to "00000".
            raise_errors (bool): True/False based on if you want the method to raise its own errors or not. Defaults to False.

        Returns:
            dict[str, str]: A dictionary of the errors the validation wants to raise.
        """
        self._validate_params_action(tableid)
        validator = StatbankUttrekksBeskrivelse(
            tableid=tableid,
            raise_errors=raise_errors,
            headers=self.__headers,
        )
        validation_errors = validator.validate(dfs)
        self.log.append(
            f"Validated data for tableid {tableid} at {(dt.datetime.now().astimezone(OSLO_TIMEZONE) + dt.timedelta(hours=1)).isoformat('T', 'seconds')}",
        )
        return validation_errors

    def transfer(
        self,
        dfs: dict[str, pd.DataFrame],
        tableid: str = "00000",
    ) -> StatbankTransfer:
        """Transfers your data to Statbanken.

        Make sure you've set the publish-date correctly before sending.

        Args:
            dfs (dict[str, pd.DataFrame]): The data to validate in a dictionary of deltabell-names as keys and pandas-dataframes as values.
            tableid (str): The tableid of the "hovedtabell" in statbanken, a 5 digit string.

        Returns:
            StatbankTransfer: An instance of the class StatbankTransfer, which details the content of a successful transfer.
        """
        self._validate_params_action(tableid)
        self.log.append(
            f"Transferring tableid {tableid} at {(dt.datetime.now().astimezone(OSLO_TIMEZONE) + dt.timedelta(hours=1)).isoformat('T', 'seconds')}",
        )
        return StatbankTransfer(
            dfs,
            tableid=tableid,
            headers=self.__headers,
            shortuser=self.shortuser,
            date=self.date,
            cc=self.cc,
            bcc=self.bcc,
            overwrite=self.overwrite,
            approve=self.approve,
        )

    @staticmethod
    def read_transfer_json(json_path_or_str: str) -> StatbankTransfer:
        """Checks if provided string exists on disk, if it does, tries to load it as json.

        Otherwise expects you to provide a json-string that works for json.loads.
        Inserts first layer in json as attributes under a blank StatbankTransfer-object.

        Args:
            json_path_or_str (str): Either a path on local storage, or a loaded json-string

        Returns:
            StatbankTransfer: An instance of the class StatbankTransfer, missing the data transferred and some other bits probably.
        """
        content = json_path_or_str
        try:
            try_path = json_path_or_str
            if Path(try_path).exists():
                with Path(try_path).open("r") as json_file:
                    content = json_file.read()
        except OSError as e:
            logger.debug(
                "Assuming you sent a json-string to open as transfer, cause that path does not exist. %s",
                str(e),
            )
        new = StatbankTransfer.__new__(StatbankTransfer)
        for k, v in json.loads(content).items():
            setattr(new, k, v)
        return new

    @staticmethod
    def apidata(
        id_or_url: str = "",
        payload: QueryWholeType | None = None,
        include_id: bool = False,
    ) -> pd.DataFrame:
        """Get the contents of a published statbank-table as a pandas Dataframe, specifying a query to limit the return.

        Args:
            id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
            payload (dict[str, str]|None): a dict of the query to include with the request, can be copied from the statbank-webpage.
            include_id (bool): If you want to include "codes" in the dataframe, set this to True

        Returns:
            pd.DataFrame: A pandas dataframe with the table-content
        """
        replace_payload: QueryWholeType = {
            "query": [],
            "response": {"format": "json-stat2"},
        }
        if payload is None:
            payload = replace_payload
        return apidata(id_or_url=id_or_url, payload=payload, include_id=include_id)

    @staticmethod
    def apidata_all(id_or_url: str = "", include_id: bool = False) -> pd.DataFrame:
        """Get ALL the contents of a published statbank-table as a pandas Dataframe.

        Args:
            id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
            include_id (bool): If you want to include "codes" in the dataframe, set this to True

        Returns:
            pd.DataFrame: A pandas dataframe with the table-content
        """
        return apidata_all(id_or_url=id_or_url, include_id=include_id)

    @staticmethod
    def apimetadata(id_or_url: str = "") -> dict[str, Any]:
        """Get the metadata of a published statbank-table as a dict.

        Args:
            id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".

        Returns:
            dict[str, Any]: The metadata of the table as the json returned from the API-get-request.
        """
        return apimetadata(id_or_url=id_or_url)

    @staticmethod
    def apicodelist(
        id_or_url: str = "",
        codelist_name: str = "",
    ) -> dict[str, str] | dict[str, dict[str, str]]:
        """Get one specific or all the codelists of a published statbank-table as a dict or nested dicts.

        Args:
            id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
            codelist_name (str): The name of the specific codelist to get.

        Returns:
            dict[str, str] | dict[str, dict[str, str]]: The codelist of the table as a dict or a nested dict.
        """
        return apicodelist(id_or_url=id_or_url, codelist_name=codelist_name)

    @staticmethod
    def apidata_rotate(
        df: pd.DataFrame,
        ind: str = "year",
        val: str = "value",
    ) -> pd.DataFrame:
        """Rotate the dataframe so that time is used as the index.

        Args:
            df (pd.dataframe): dataframe (from <get_from_ssb> function
            ind (str): string of column name denoting time
            val (str): string of column name denoting values

        Returns:
            pd.DataFrame: pivoted dataframe
        """
        return apidata_rotate(df, ind, val)

    def _validate_date(self) -> None:
        """Validate dates provided to the client."""
        if not (isinstance(self.date, (dt.date, dt.datetime))):
            error_msg = "Date must be a datetime.datetime or datetime.date"  # type: ignore[unreachable]
            raise TypeError(error_msg)
        # Date should not be on a weekend
        if self.date.weekday() in [5, 6]:
            logger.warning(
                "Warning, you are publishing during a weekend, this is not common practice.",
            )

    # Class meta-validation
    def _validate_params_action(self, tableid: str) -> None:
        """Validates tableid mainly, more actively than other params."""
        if not isinstance(tableid, str):
            error_msg = f"{tableid} is not a string."  # type: ignore[unreachable]
            raise TypeError(error_msg)
        if (
            tableid.isdigit() and len(tableid) != STATBANK_TABLE_ID_LEN
        ):  # Allow for "hovednavn" in addition to tableid
            error_msg = f"{tableid} is numeric, but not 5 characters long."
            raise ValueError(error_msg)

    def _validate_params_init(self) -> None:
        """Validates many of the parameters sent in on client-initialization."""
        if not self.shortuser:
            self.shortuser = self._get_user_tbf()
        if not self.cc:
            self.cc = self.shortuser
        if not self.bcc:
            self.bcc = self.cc
        if not isinstance(self.overwrite, bool):
            error_msg = "(Bool) Set overwrite to either False = no overwrite (dublicates give errors), or  True = automatic overwrite"  # type: ignore[unreachable]
            raise TypeError(error_msg)
        if not isinstance(self.approve, int) or self.approve not in iter(Approve):
            error_msg = "(Approve) Set approve to either 0 = manual, 1 = automatic (immediatly), or 2 = JIT-automatic (just-in-time)"
            raise ValueError(error_msg)

    @staticmethod
    def _get_user_tbf() -> str:
        user_mail = os.environ.get("GIT_USER_MAIL", "")
        if not user_mail:
            user_mail = os.environ.get("JUPYTERHUB_USER", "")
        if "@" in user_mail:
            user_mail = user_mail.split("@")[0]
        return user_mail



"""Validates and transfers data from Dapla to Statbank. Gets data from public and internal statbank.

Used internally by SSB (Statistics Norway).
"""

from __future__ import annotations

import datetime as dt
import importlib
import importlib.metadata  # Needed even with whole import over

import toml

from statbank.apidata import apicodelist
from statbank.apidata import apidata
from statbank.apidata import apidata_all
from statbank.apidata import apidata_rotate
from statbank.apidata import apimetadata
from statbank.client import StatbankClient
from statbank.statbank_logger import logger

__all__ = [
    "StatbankClient",
    "apidata",
    "apidata_all",
    "apidata_rotate",
    "apimetadata",
    "apicodelist",
]


# Split into function for testing
def _try_getting_pyproject_toml(e: Exception | None = None) -> str:
    if e is None:
        passed_excep: Exception = Exception("")
    else:
        passed_excep = e
    try:
        try:
            version: str = toml.load("../pyproject.toml")["tool"]["poetry"]["version"]
        except FileNotFoundError:
            version = toml.load("./pyproject.toml")["tool"]["poetry"]["version"]
    except toml.TomlDecodeError as e:
        version = "0.0.0"
        logger.exception(
            "Error from dapla-statbank-clients __init__, not able to get version-number, setting it to %s. Exception: %s",
            version,
            str(passed_excep),
        )
    return version


# Gets the installed version from pyproject.toml, then there is no need to update this file
try:
    __version__ = importlib.metadata.version("dapla-statbank-client")
except importlib.metadata.PackageNotFoundError as e:
    __version__ = _try_getting_pyproject_toml(e)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from statbank.api_types import DelTabellType
    from statbank.api_types import KodelisteTypeParsed
    from statbank.api_types import KolonneStatistikkvariabelType
    from statbank.api_types import KolonneVariabelType
    from statbank.api_types import SuppressionCodeListType

import pandas as pd

from statbank.statbank_logger import logger


class StatbankValidateError(Exception):
    """Use when raising errors stemming from the validators not running cleanly."""


class StatbankUttrekkValidators:
    """Split out from the main Uttrekk-class, this class contains all the validator-methods."""

    def __init__(self) -> None:
        """This init will never be used directly, as this class is always inherited from.

        So, these attribute-settings are for type-checking with mypy.
        """
        self.subtables: dict[str, str] = {}
        self.variables: list[DelTabellType] = []
        self.codelists: dict[str, KodelisteTypeParsed] = {}
        self.suppression: None | list[SuppressionCodeListType] = None

    def _validate_number_dataframes(self, data: dict[str, pd.DataFrame]) -> None:
        # Number subtables should match length of data-iterable
        if len(self.subtables.values()) != len(data.values()):
            error_msg = f"""Please put one or more pandas Dataframes in a dict as your data. Keys in the dict should be "deltabell-navn": {self.subtables.keys()}"""
            raise TypeError(error_msg)
        for k, df in data.items():
            if not isinstance(df, pd.DataFrame):
                error_msg = f"{k}'s value is not a dataframe"  # type: ignore[unreachable]
                raise TypeError(error_msg)

    def _validate_number_columns(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        # Number of columns in data must match beskrivelse
        for deltabell_num, deltabell in enumerate(self.variables):
            deltabell_navn = deltabell["deltabell"]
            col_num = len(deltabell["variabler"]) + len(
                deltabell["statistikkvariabler"],
            )  # Mangler prikke-kolonner?
            if "null_prikk_missing" in deltabell:
                col_num += len(deltabell["null_prikk_missing"])
            if "internasjonal_rapportering" in deltabell:
                col_num += len(deltabell["internasjonal_rapportering"])
            if len(data[deltabell_navn].columns) != col_num:
                error_msg = f"""
                            EXPECTING {col_num} COLUMNS IN DATAFRAME NUMBER {deltabell_num}:
                            {deltabell_navn}
                            BUT FOUND {len(data[deltabell_navn].columns)}
                            """
                validation_errors[f"col_count_data_{deltabell_num}"] = ValueError(
                    error_msg,
                )
        for k in validation_errors:
            if "col_count_data" in k:
                logger.warning(validation_errors[k])
                break
        else:
            logger.debug("Correct number of columns...")
        return validation_errors

    def _check_for_literal_nans_in_strings(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        for name, df in data.items():
            string_df = df.select_dtypes(
                include=["object", "string", "string[pyarrow]"],
            )
            cat_df = df.select_dtypes(include=["category"])

            nans = ["nan", "na", "none", "."]
            validation_errors = self._check_for_literal_nans_string_dtype(
                string_df,
                name,
                nans,
                validation_errors,
            )
            validation_errors = self._check_for_literal_nans_categorical_dtype(
                cat_df,
                name,
                nans,
                validation_errors,
            )
        return validation_errors

    @staticmethod
    def _check_for_literal_nans_string_dtype(
        string_df: pd.DataFrame,
        name: str,
        nans: list[str],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        if len(string_df.columns):
            for col in string_df.columns:
                error_text = f"""{col} in {name} has strings, that look like NAs / empty cells,
                (In this list: {nans})
                Which have been converted to literal strings.
                Consider handeling your NAs before converting them to strings.
                Maybe with a .fillna("") before an .astype(str) """
                nan_len = len(string_df[string_df[col].str.lower().isin(nans)])
                if nan_len:
                    validation_errors[f"contains_string_nans_{name}_{col}"] = (
                        ValueError(error_text)
                    )
                    logger.warning(error_text)
        return validation_errors

    @staticmethod
    def _check_for_literal_nans_categorical_dtype(
        cat_df: pd.DataFrame,
        name: str,
        nans: list[str],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        if len(cat_df.columns):
            for col in cat_df.columns:
                error_text = f"""{col} in {name} is a categorical but has strings,
                that look like NAs / empty cells,
                (In this list: {nans})
                Which have been converted to literal strings?
                Consider handeling your NAs before converting them to strings.
                Maybe with a .fillna("") before an .astype(str) """
                nan_cats = [
                    cat for cat in cat_df[col].cat.categories if cat.lower() in nans
                ]
                if nan_cats:
                    validation_errors[
                        f"contains_string_nans_in_category_{name}_{col}"
                    ] = ValueError(error_text)
                    logger.warning(error_text)
        return validation_errors

    def _check_for_floats(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        for name, df in data.items():
            for col in df.columns:
                if "float" in str(df[col].dtype).lower():
                    error_text = f"""{col} in {name} is a float.
                    Consider running the dict of dataframes through:
                    data = uttrekksbeskrivelse.round_data(data),
                    this rounds UP like SAS and Excel, not to-even as
                    Python does otherwise."""
                    validation_errors[f"contains_floats_{name}_{col}"] = ValueError(
                        error_text,
                    )
                    logger.warning(error_text)
        return validation_errors

    def _check_time_formats(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        # Time-columns should follow time format
        for deltabell in self.variables:
            for variabel in deltabell["variabler"]:
                if "Kodeliste_text" in variabel and "format = " in variabel.get(
                    "Kodeliste_text",
                    "",
                ):
                    validation_errors = self._check_time_columns(
                        deltabell["deltabell"],
                        variabel,
                        data,
                        validation_errors,
                    )
        for k in validation_errors:
            matches = [
                "time_non_digit_column",
                "character_match_column",
                "special_character_match_column",
                "time_single_length_format",
                "time_formatlength",
            ]
            if [x for x in matches if x in k]:
                break
        else:
            logger.debug("Timeformat validation ok.")

        return validation_errors

    def _check_time_same_values_in_deltabeller(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        # Time-columns should have same values in all deltabeller

        for deltabell in self.variables:
            for variabel in deltabell["variabler"]:
                if "Kodeliste_text" in variabel and "format = " in variabel.get(
                    "Kodeliste_text",
                    "",
                ):
                    validation_errors = (
                        self._check_time_same_values_in_deltabeller_data(
                            deltabell,
                            variabel,
                            data,
                            validation_errors,
                        )
                    )

        return validation_errors

    @staticmethod
    def _check_time_same_values_in_deltabeller_data(
        deltabell: DelTabellType,
        variabel: KolonneVariabelType,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        found: list[str] = []
        col_num = int(variabel["kolonnenummer"]) - 1
        for del_data in data.values():
            unique_times = [str(x) for x in del_data.iloc[:, col_num].unique()]
            if found == []:
                found = unique_times
            else:
                times_missing = [x for x in found if x not in unique_times]
                times_extra = [x for x in unique_times if x not in found]
                if times_missing:
                    validation_errors[
                        f"time_missing_in_deltabell_{deltabell['deltabell']}"
                    ] = ValueError(
                        f"""Time(s) {times_missing} missing in deltabell {deltabell['deltabell']}""",
                    )
                if times_extra:
                    validation_errors[
                        f"time_extra_in_deltabell_{deltabell['deltabell']}"
                    ] = ValueError(
                        f"""Time(s) {times_extra} extra in deltabell {deltabell['deltabell']}""",
                    )
        return validation_errors

    def _check_time_columns(
        self,
        deltabell_name: str,
        variabel: KolonneVariabelType,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        col_num = int(variabel["kolonnenummer"]) - 1
        timeformat_raw = (
            variabel["Kodeliste_text"].split(" format = ")[1].strip().replace("Å", "å")
        )
        # Check length of coloumn matches length of format
        if (
            len(data[deltabell_name].iloc[:, col_num].astype(str).str.len().unique())
            != 1
        ):
            validation_errors[f"time_single_length_format_{col_num}"] = ValueError(
                f"""Column number {col_num} does not have
                a single time format
                in the shape: {timeformat_raw}""",
            )
        if (
            len(timeformat_raw)
            != data[deltabell_name].iloc[:, col_num].astype(str).str.len().unique()[0]
        ):
            validation_errors[f"time_formatlength_{col_num}"] = ValueError(
                f"""Column number {col_num} does not match
                time format in the shape: {timeformat_raw}""",
            )

        validation_errors = self._check_time_nums(
            timeformat_raw,
            deltabell_name,
            col_num,
            data,
            validation_errors,
        )
        validation_errors = self._check_time_chars(
            timeformat_raw,
            deltabell_name,
            col_num,
            data,
            validation_errors,
        )
        return self._check_time_specials(
            timeformat_raw,
            deltabell_name,
            col_num,
            data,
            validation_errors,
        )

    @staticmethod
    def _check_time_nums(
        timeformat_raw: str,
        deltabell_name: str,
        col_num: int,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        nums: list[int] = [i for i, c in enumerate(timeformat_raw) if c.islower()]
        if nums:
            for num in nums:
                if not all(
                    data[deltabell_name].iloc[:, col_num].str[num].str.isdigit(),
                ):
                    validation_errors[f"time_non_digit_column{col_num}"] = ValueError(
                        f"Character number {num} in column {col_num} in DataFrame {deltabell_name}, does not match format {timeformat_raw}",
                    )
        return validation_errors

    @staticmethod
    def _check_time_chars(
        timeformat_raw: str,
        deltabell_name: str,
        col_num: int,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        chars: dict[int, str] = {
            i: c for i, c in enumerate(timeformat_raw) if c.isupper()
        }
        if chars:
            for i, char in chars.items():
                if not all(data[deltabell_name].iloc[:, col_num].str[i] == char):
                    validation_errors[f"character_match_column{col_num}"] = ValueError(
                        f"Should be capitalized character? Character {char}, character number {i} in column {col_num} in DataFrame {deltabell_name}, does not match format {timeformat_raw}",
                    )
        return validation_errors

    @staticmethod
    def _check_time_specials(
        timeformat_raw: str,
        deltabell_name: str,
        col_num: int,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        specials: dict[int, str] = {
            i: c for i, c in enumerate(timeformat_raw) if not c.isalnum()
        }
        if specials:
            for i, special in specials.items():
                if not all(data[deltabell_name].iloc[:, col_num].str[i] == special):
                    validation_errors[f"special_character_match_column{col_num}"] = (
                        ValueError(
                            f"Should be the special character {special}, character number {i} in column {col_num} in DataFrame {deltabell_name}, does not match format {timeformat_raw}",
                        )
                    )
        return validation_errors

    def _check_suppression(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        if self.suppression:
            prikk_codes = [code["Kode"] for code in self.suppression]
            prikk_codes += [""]
            validation_errors = self._check_suppression_variables(
                data,
                prikk_codes,
                validation_errors,
            )
        for k in validation_errors:
            if "prikke_character_match_column" in k:
                break
        else:
            logger.debug(
                "suppression-codes validation ok / No prikke-columns in use.",
            )

        return validation_errors

    def _check_suppression_variables(
        self,
        data: dict[str, pd.DataFrame],
        prikk_codes: list[str],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        for deltabell in self.variables:
            deltabell_name = deltabell["deltabell"]
            if "null_prikk_missing" in deltabell:
                for prikk_col in deltabell["null_prikk_missing"]:
                    col_num = int(prikk_col["kolonnenummer"]) - 1
                    if not all(
                        data[deltabell_name].iloc[:, col_num].isin(prikk_codes),
                    ):
                        validation_errors[f"prikke_character_match_column{col_num}"] = (
                            ValueError(
                                f"Prikke-code not among allowed prikkecodes: {prikk_codes}, in column {col_num} in DataFrame {deltabell_name}.",
                            )
                        )
        return validation_errors

    def _check_unique_combinations_categories_times(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        # Get column-numbers containing categorical values per deltabell
        for deltabell in self.variables:
            category_col_nums = [
                int(var["kolonnenummer"]) - 1 for var in deltabell["variabler"]
            ]
            df_colcheck = data[deltabell["deltabell"]].iloc[:, category_col_nums]
            if df_colcheck.duplicated().any():
                validation_errors[
                    f"duplicate_categorical_time_groups_{deltabell['deltabell']}"
                ] = ValueError(
                    f"There seems to be duplicate rows across the categorical values (including time) in deltabell {deltabell['deltabell']}.",
                )
        for k in validation_errors:
            if "duplicate_categorical_time_groups" in k:
                break
        else:
            logger.debug(
                "Found no duplicate combinations of categorical columns",
            )

        return validation_errors

    def _get_check_codes(self) -> dict[str, dict[str, list[str]]]:
        check_codes: dict[str, dict[str, list[str]]] = {}
        for deltabell in self.variables:
            deltabell_navn = deltabell["deltabell"]
            check_codes[deltabell_navn] = {}
            for variabel in deltabell["variabler"]:
                if (
                    "Kodeliste_id" in variabel
                    and variabel.get("Kodeliste_id", "") != "-"
                ):
                    check_codes[deltabell_navn][variabel["kolonnenummer"]] = list(
                        self.codelists[variabel["Kodeliste_id"]]["koder"].keys(),
                    )
        return check_codes

    def _category_columns_are_strings(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        categorycode_not_string = []
        check_codes = self._get_check_codes()
        for deltabell_name, variabel in check_codes.items():
            for col_num in variabel:
                col = data[deltabell_name].iloc[:, int(col_num) - 1]
                if not pd.api.types.is_string_dtype(col):
                    categorycode_not_string += [
                        f"""{col_num} is a categorical column, but is not a string columns?
                            Convert the column to string before validating / transferring.
                            {col_num}, in deltabell {deltabell_name}""",
                    ]
        if categorycode_not_string:
            logger.warning("Codes in data, outside codelist:")
            logger.warning("\n".join(categorycode_not_string))
            validation_errors["categorycode_outside"] = ValueError(
                categorycode_not_string,
            )
        else:
            logger.debug(
                "All categorical columns are string, thats correct.",
            )
        return validation_errors

    def _category_code_usage(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:

        validation_errors = self._check_category_code_usage_outside(
            data,
            validation_errors,
        )
        return self._check_category_code_usage_missing(
            data,
            validation_errors,
        )

    def _check_category_code_usage_missing(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        categorycode_missing = []
        check_codes = self._get_check_codes()
        for deltabell_name, variabel in check_codes.items():
            for col_num, codelist in variabel.items():
                col = data[deltabell_name].iloc[:, int(col_num) - 1]
                col_unique = col.unique()

                for kod in codelist:
                    if kod not in col_unique:
                        categorycode_missing += [
                            f"""Code {kod} missing from column number
                            {col_num}, in deltabell {deltabell_name}""",
                        ]
        # No values outside, warn of missing from codelists on categorical columns
        if categorycode_missing:
            logger.info(
                """Category codes missing from data (This is ok,
            just make sure missing data is intentional):""",
            )
            logger.info("\n".join(categorycode_missing))
        else:
            logger.debug("No codes missing from categorical columns.")
        return validation_errors

    def _check_category_code_usage_outside(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        check_codes = self._get_check_codes()
        categorycode_outside = []
        for deltabell_name, variabel in check_codes.items():
            for col_num, codelist in variabel.items():
                col = data[deltabell_name].iloc[:, int(col_num) - 1]
                col_unique = col.unique()
                for kod in col_unique:
                    if kod not in codelist and " " in kod:
                        categorycode_outside += [
                            f"""{kod} contains spaces, should it?
                            The exact code "{kod}" (including spaces) is in the data, but not in uttrekksbeskrivelse,
                            add to statbank admin? From column number
                            {col_num}, in deltabell {deltabell_name}""",
                        ]
                    elif kod not in codelist:
                        categorycode_outside += [
                            f"""Code {kod} in data, but not in uttrekksbeskrivelse,
                            add to statbank admin? From column number
                            {col_num}, in deltabell {deltabell_name}""",
                        ]
        if categorycode_outside:
            logger.warning("Codes in data, outside codelist:")
            logger.warning("\n".join(categorycode_outside))
            validation_errors["categorycode_outside"] = ValueError(categorycode_outside)
        else:
            logger.debug(
                "No codes in categorical columns outside codelist.",
            )
        return validation_errors

    def _check_statistikkvar_numerical(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        for deltabell in self.variables:
            deltabell_name = deltabell["deltabell"]
            for variabel in deltabell["statistikkvariabler"]:
                col_num = int(variabel["kolonnenummer"]) - 1
                col = data[deltabell_name].iloc[:, col_num].copy()
                # Check if can be converted to float and int
                try:
                    if pd.api.types.is_string_dtype(col):
                        col = col.str.replace(",", ".", regex=False)
                    col.astype("Float64")
                except ValueError as e:
                    validation_errors[
                        f"statistikkvar_not_numerical_column{col_num}"
                    ] = ValueError(e)
        return validation_errors

    def _check_rounding(
        self,
        data: dict[str, pd.DataFrame],
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        """If a column should have a set number of decimals, check if its a string, and how many places are used after the decimal seperator: ","."""
        for deltabell in self.variables:
            deltabell_name = deltabell["deltabell"]
            for variabel in deltabell["statistikkvariabler"]:
                validation_errors = self._check_rounding_per_variable(
                    variabel,
                    data,
                    deltabell_name,
                    validation_errors,
                )
        return validation_errors

    @staticmethod
    def _check_rounding_per_variable(
        variabel: KolonneStatistikkvariabelType,
        data: dict[str, pd.DataFrame],
        deltabell_name: str,
        validation_errors: dict[str, ValueError],
    ) -> dict[str, ValueError]:
        col_num = int(variabel["kolonnenummer"]) - 1
        decimal_num = int(variabel["Antall_lagrede_desimaler"])
        error = False
        column = (
            data[deltabell_name]
            .iloc[:, col_num]
            .copy()
            .astype(str)
            .str.replace(".", ",", regex=False)
        )
        # Compensate for na / empty cells
        column = column[column != ""]
        # If there are no rows left after removing empty, skip checking more
        if not len(column):
            return validation_errors
        if decimal_num:
            col_decimals = column.str.split(",").str[-1].str.len()
            if (col_decimals != decimal_num).any():
                error = True
        elif not (
            column.str.replace("-", "", regex=False)
            .str.replace(",", "", regex=False)
            .str.replace(".", "", regex=False)
            .str.isdigit()
            .all()
        ):
            error = True

        if error:
            error_text = f"""Check that string column {col_num} in
            {deltabell_name} that should be rounded to
            {decimal_num} decimal places,
            has correct number of decimals. And consider converting from a
            non-rounded float to a string with this method:
            data = uttrekksbeskrivelse.round_data(data),
        this rounds UP like SAS and Excel, not to-even as
        Python does otherwise."""
            validation_errors[f"rounding_error_{deltabell_name}_{col_num}"] = (
                ValueError(error_text)
            )
            logger.warning(error_text)
        return validation_errors



from __future__ import annotations

import gc
import json
import os
import urllib
from datetime import datetime as dt
from datetime import timedelta as td
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
import requests as r

from statbank.auth import StatbankAuth
from statbank.globals import APPROVE_DEFAULT_JIT
from statbank.globals import OSLO_TIMEZONE
from statbank.globals import SSB_TBF_LEN
from statbank.globals import Approve
from statbank.globals import _approve_type_check
from statbank.statbank_logger import logger

if TYPE_CHECKING:
    from statbank.api_types import TransferResultType


class StatbankTransfer(StatbankAuth):
    """Class for talking with the "transfer-API", which actually recieves the data from the user and sends it to Statbank.

    Attributes:
        data (dict[str, pd.DataFrame]):  name of "deltabell.dat" as keys.
            Number of DataFrames needs to match the number of "deltabeller" in
            the uttakksbeskrivelse.
            Dict-shape can be retrieved and validated before transfer with the
            Uttakksbeskrivelses-class.
        tableid (str): The numeric id of the table, matching the one found on the website.
            Should be a 5-length numeric-string. Alternatively it should be possible to send in the "hovedtabellnavn" instead of the tableid.
        shortuser (str): The abbrivation of username at ssb. Three letters, like "cfc"
        date (str): Date for publishing the transfer. Shape should be "yyyy-mm-dd", like "2022-01-01".
            Statbanken only allows publishing four months into the future?
        cc (str): First person to be notified by email of transfer. Defaults to the same as "shortuser"
        bcc (str): Second person to be notified by email of transfer. Defaults to the same as "cc"
        overwrite (bool):
            - False = no overwrite
            - True = overwrite
        approve (Approve | str | int):
            - 0 = MANUAL approval
            - 1 = AUTOMATIC approval at transfer-time (immediately)
            - 2 = JIT (Just In Time), approval right before publishing time
        validation (bool):
            - True, if you want the python-validation code to run user-side.
            - False, if its slow and unnecessary.
        boundary (str): String that defines the splitting of the body in the transfer-post-request.
            Kept here for uniform choice through the class.
        urls (dict[str, str]): Urls for transfer, observing the result etc.,
            built from environment variables.
        headers (dict[str, str]): Might be deleted without warning.
            Temporarily holds the Authentication for the request.
        params (dict[str, str]): This dict will be built into the post request.
            Keep it in this nice shape for later introspection.
        body (str): The data parsed into the body-shape the Statbank-API expects in the transfer-post-request.
        response (requests.Response): The resulting response from the transfer-request. Headers might be deleted without warning.
    """

    def __init__(  # noqa: PLR0913
        self,
        data: dict[str, pd.DataFrame],
        tableid: str = "",
        shortuser: str = "",
        date: dt | str | None = None,
        cc: str = "",
        bcc: str = "",
        overwrite: bool = True,
        approve: int | str | Approve = APPROVE_DEFAULT_JIT,
        validation: bool = True,
        delay: bool = False,
        headers: dict[str, str] | None = None,
    ) -> None:
        """Make the transfer to statbanken at the end of initializing the object.

        May run the validations from the StatbankValidation class before the transfer.
        """
        self._set_user_attrs(shortuser=shortuser, cc=cc, bcc=bcc)
        self._set_date(date=date)
        self.data = data
        if not isinstance(tableid, str) or not tableid.isdigit():
            error_msg = "Loaduser is no longer a parameter, make sure the tableid parameter is a string of digits."
            raise ValueError(error_msg)
        self.tableid = tableid
        self.overwrite = overwrite
        self.approve = _approve_type_check(approve)
        self.validation = validation
        self.__delay = delay
        self.oppdragsnummer: str = ""
        self.boundary = "12345"
        if validation:
            self._validate_original_parameters()

        self.urls = self._build_urls()
        if not self.delay:
            if headers:
                self.transfer(headers)
            else:
                self.transfer()

    def transfer(self, headers: dict[str, str] | None = None) -> None:
        """Transfers your data to Statbanken.

        Make sure you've set the publish-date correctly before sending.
        Will only work if the transfer has not already been sent, meaning it was "delayed".

        Args:
            headers (dict[str, str] | None): Mostly for internal use by the package.
                Needs to be a finished compiled headers for a request including Authorization.

        Raises:
            ValueError: If the transfer is already transferred.
        """
        # In case transfer has already happened, don't transfer again
        if self.oppdragsnummer:
            error_msg = f"Already transferred? {self.urls['gui'] + self.oppdragsnummer} Remake the StatbankTransfer-object if intentional."
            raise ValueError(error_msg)
        if headers is None:
            self.headers = self._build_headers()
        else:
            self.headers = headers
        try:
            self.params = self._build_params()
            self._validate_datatype()
            self.body = self._body_from_data()

            url_load_params = self.urls["loader"] + urllib.parse.urlencode(self.params)
            urllib.parse.urlparse(url_load_params)  # Test to see if url is valid format
            self.response = self._make_transfer_request(url_load_params)
            self._cleanup_response()
        finally:
            del self.headers  # Cleaning up auth-storing
            self.__delay = False
        self._handle_response()

    def __str__(self) -> str:
        """Print a string with the status of the transfer."""
        first_line = f"Overføring for statbanktabell {self.tableid}.\n"
        if self.delay:
            result = f"""{first_line}Ikke overført enda."""
        else:
            result = f"""{first_line}Publisering: {self.date}.\nLastelogg: {self.urls['gui'] + self.oppdragsnummer}"""
        return result

    def __repr__(self) -> str:
        """Get a representation of how to recreate the object using parameters."""
        return f'StatbankTransfer([data], tableid="{self.tableid}")'

    @property
    def delay(self) -> bool:
        """Obfuscate the delay a bit from the user. We dont want transfers transferring again without recreating the object."""
        return self.__delay

    def _set_user_attrs(
        self,
        shortuser: str = "",
        cc: str = "",
        bcc: str = "",
    ) -> None:
        if shortuser:
            self.shortuser = shortuser
        else:
            self.shortuser = os.environ["JUPYTERHUB_USER"].split("@")[0]
        if cc:
            self.cc = cc
        else:
            self.cc = self.shortuser
        if bcc:
            self.bcc = bcc
        else:
            self.bcc = self.cc

    def _set_date(self, date: dt | str | None = None) -> None:
        # At this point we want date to be a string?
        if date is None:
            date = dt.now().astimezone(OSLO_TIMEZONE) + td(days=1, hours=1)
        if isinstance(date, str):
            self.date: str = date
        else:
            self.date = date.strftime("%Y-%m-%d")

    def to_json(self, path: str = "") -> str | None:
        """Store a copy of the current state of the transfer-object as a json.

        If path is provided, tries to write to it,
        otherwise will return a json-string for you to handle like you wish.

        Args:
            path (str): if provided, will try to write a json to a local path.

        Returns:
            None: If path is provided, tries to write a json there and returns nothing.
            str: If path is not provided, returns the json-string for you to handle as you wish.
        """
        logger.warning(
            "Warning, some nested, deeper data-structures"
            " like dataframes and other class-objects will not be serialized",
        )
        json_content = json.dumps(self.__dict__, default=lambda _: "<not serializable>")
        # If path provided write to it, otherwise return the string-content
        if path:
            logger.info("Writing to %s", path)
            with Path(path).open(mode="w") as json_file:
                json_file.write(json_content)
        else:
            return json.dumps(json_content)
        return None

    def _validate_original_parameters(self) -> None:
        for _, shortuser in enumerate([self.shortuser, self.cc, self.bcc]):
            if len(shortuser) != SSB_TBF_LEN or not isinstance(shortuser, str):
                error_msg = f'Brukeren {shortuser} - "trebokstavsforkortelse" - må være tre bokstaver...'
                raise ValueError(error_msg)

        if not isinstance(self.date, str) or not self._valid_date_form(self.date):
            error_msg = "Skriv inn datoformen for publisering som 1900-01-01"
            raise TypeError(error_msg)

        if not isinstance(self.overwrite, bool):
            error_msg = "(Bool) Sett overwrite til enten False = ingen overskriving (dubletter gir feil), eller  True = automatisk overskriving."  # type: ignore[unreachable]
            raise TypeError(error_msg)

        if self.approve not in iter(Approve):
            error_msg = "(Integer) Sett approve til enten 0 = manuell, 1 = automatisk (umiddelbart), eller 2 = JIT-automatisk (just-in-time)"
            raise ValueError(error_msg)

    def _validate_datatype(self) -> None:
        for deltabell_name, deltabell_data in self.data.items():
            if not isinstance(deltabell_name, str):
                error_msg = f"{deltabell_name} is not a string."  # type: ignore[unreachable]
                raise TypeError(error_msg)
            if not isinstance(deltabell_data, pd.DataFrame):
                error_msg = f"Data for {deltabell_name}, must be a pandas DataFrame"  # type: ignore[unreachable]
                raise TypeError(error_msg)

    def _body_from_data(self) -> str:
        # Data should be a iterable of pd.DataFrames at this point,
        # reshape to body
        body = ""
        for filename, elem in self.data.items():
            # Replace all nans in data
            elem_fillna = elem.copy().fillna("")
            body += f"--{self.boundary}"
            body += f"\nContent-Disposition:form-data; filename={filename}"
            body += "\nContent-type:text/plain\n\n"
            csv_content = elem_fillna.to_csv(sep=";", index=False, header=False)
            body += str(csv_content)
        body += f"\n--{self.boundary}--"
        return body.replace("\n", "\r\n")  # Statbank likes this?

    @staticmethod
    def _valid_date_form(date: str) -> bool:
        if (date[:4] + date[5:7] + date[8:]).isdigit() and (
            (date[4] + date[7]) == "--"
        ):
            return True
        return False

    def _build_params(self) -> dict[str, str | int]:
        if isinstance(self.date, dt):  # type: ignore[unreachable]
            date = self.date.strftime("%Y-%m-%d")  # type: ignore[unreachable]
        else:
            date = self.date
        return {
            "initialier": self.shortuser,
            "hovedtabell": self.tableid,
            "publiseringsdato": date,
            "fagansvarlig1": self.cc,
            "fagansvarlig2": self.bcc,
            "auto_overskriv_data": str(int(self.overwrite)),
            "auto_godkjenn_data": str(int(self.approve)),
        }

    def _make_transfer_request(
        self,
        url_params: str,
    ) -> r.Response:
        result = r.post(url_params, headers=self.headers, data=self.body, timeout=15)
        # Trying to clean all auth etc out of response
        result.raise_for_status()
        return result

    def _cleanup_response(self) -> None:
        if hasattr(self.response.request, "headers"):
            del (
                self.response.request.headers
            )  # Auth is stored here also, for some reason
        if hasattr(self.response, "cookies"):
            del self.response.cookies
        if hasattr(self.response, "raw"):
            del self.response.raw
        gc.collect()  # Hoping this removes the del-ed stuff from memory

    def _handle_response(self) -> None:
        resp_json: TransferResultType = self.response.json()
        response_msg = resp_json["TotalResult"]["Message"]
        self.oppdragsnummer = response_msg.split("lasteoppdragsnummer:")[1].split(" =")[
            0
        ]
        if not self.oppdragsnummer.isdigit():
            error_msg = (
                f"Lasteoppdragsnummer: {self.oppdragsnummer} er ikke ett rent nummer."
            )
            raise ValueError(error_msg)

        publish_date = dt.strptime(
            response_msg.split("Publiseringsdato '")[1].split("',")[0],
            "%d.%m.%Y %H:%M:%S",
        ).astimezone(OSLO_TIMEZONE) + td(hours=1)
        publish_hour = int(response_msg.split("Publiseringstid '")[1].split(":")[0])
        publish_minute = int(
            response_msg.split("Publiseringstid '")[1].split(":")[1].split("'")[0],
        )
        publish_time = publish_hour * 3600 + publish_minute * 60
        publish_date = publish_date + td(0, publish_time)
        logger.info("Publisering satt til: %s", publish_date.isoformat("T", "seconds"))
        logger.info(
            "Følg med på lasteloggen (tar noen minutter): %s",
            {self.urls["gui"] + self.oppdragsnummer},
        )
        self.resp_json = resp_json



from __future__ import annotations

import datetime as dt
import enum


class Approve(enum.IntEnum):
    """Enum for approval codes."""

    MANUAL = 0
    """Manual approval."""
    AUTOMATIC = 1
    """Automatic approval at transfer-time (immediately)."""
    JIT = 2
    """Just in time approval right before publishing time."""


def _approve_type_check(approve: Approve | int | str) -> Approve:
    if isinstance(approve, int) and not isinstance(approve, Approve):
        result: Approve = Approve(approve)
    elif isinstance(approve, str) and approve.isdigit():
        result = Approve(int(approve))
    elif isinstance(approve, str):
        result = getattr(Approve, approve)
    elif isinstance(approve, Approve):
        result = approve
    else:
        error_msg = f"Dont know how to handle approve of type {type(approve)}"  # type: ignore[unreachable]
        raise TypeError(error_msg)
    return result


OSLO_TIMEZONE = dt.timezone(dt.timedelta(hours=1))
TOMORROW = dt.datetime.now(tz=OSLO_TIMEZONE) + dt.timedelta(days=1)
APPROVE_DEFAULT_JIT = Approve.JIT
STATBANK_TABLE_ID_LEN = 5
REQUEST_OK = 200
SSB_TBF_LEN = 3



from __future__ import annotations

from typing import TypedDict

from typing_extensions import NotRequired

#################################
# Statbank get public data type #
#################################


class QueryWholeType(TypedDict):
    """The whole query-type is used to get the data from Statbank."""

    query: list[QueryPartType]
    response: dict[str, str]


class QueryPartType(TypedDict):
    """This represents each part of the query sent to statbanken."""

    code: str
    selection: SelectionPartType


class SelectionPartType(TypedDict):
    """The selection-type is used to get the data from Statbank."""

    filter: str
    values: list[str]


####################
# TRANSFER RESULTS #
####################


class TransferPartType(TypedDict):
    """There are multiple copies of this for every json."""

    GeneratedId: None | str
    Status: str
    Message: str
    Exception: None | str
    ValidationInfoItems: None | str


class TransferResultType(TypedDict):
    """The types of the main result-json from a transfer."""

    TotalResult: TransferPartType
    ItemResults: list[TransferPartType]


#######################
# UTTREKKSBESKRIVELSE #
#######################


class SuppressionCodeListType(TypedDict):
    """The suppression-codes are used to filter out rows in the Uttrekksbeskrivelse, lower-level."""

    Kode: str
    Vises_som: str
    Beskrivelse: str


class SuppressionDeltabellCodeListType(TypedDict):
    """The suppression-codes are used to filter out rows in the Uttrekksbeskrivelse, top-level."""

    kolonnenummer: str
    gjelder_for_text: str
    gjelder_for__kolonner_nummer: str


class KodeType(TypedDict):
    """The code -> labels for the list in the Codelists."""

    kode: str
    text: str


class KodelisteType(TypedDict):
    """The codelists have their own name, which links from the column-metadata, and a list of code -> labels."""

    kodeliste: str
    SumIALtTotalKode: NotRequired[str]
    koder: list[KodeType]


class KodelisteTypeParsed(TypedDict):
    """Reorganizing the kodelists to this is done under init of Uttrekk."""

    SumIALtTotalKode: NotRequired[str]
    koder: dict[str, str]


class KolonneInternasjonalRapporteringType(TypedDict):
    """The international-report-variables are rarely used, but has other metadata than the other columns.."""

    kolonnenummer: str
    egenskap: str
    beskrivning: str
    Kodeliste_id: str
    Kodeliste_text: str


class KolonneStatistikkvariabelType(TypedDict):
    """The statistic-variables are mostly the numerical variables in statbanken.

    This class describes the types of the metadata about these columns in the Uttrekksbeskrivelse.
    """

    kolonnenummer: str
    Text: str
    Enhet: str
    Antall_lagrede_desimaler: str
    Antall_viste_desimaler: str


class KolonneVariabelType(TypedDict):
    """The Variables are mostly the classification variables in statbanken.

    This class describes the types of the metadata about these columns in the Uttrekksbeskrivelse.
    """

    kolonnenummer: str
    Klassifikasjonsvariabel: str
    Variabeltext: str
    Kodeliste_id: str
    Kodeliste_text: str


class DelTabellType(TypedDict):
    """A typed subpart of the returned Filbeskrivelse."""

    deltabell: str
    variabler: list[KolonneVariabelType]
    statistikkvariabler: list[KolonneStatistikkvariabelType]
    internasjonal_rapportering: NotRequired[list[KolonneInternasjonalRapporteringType]]
    null_prikk_missing: NotRequired[list[SuppressionDeltabellCodeListType]]
    eksempel_linje: str


class FilBeskrivelseType(TypedDict):
    """The content of the json returned by the uttrekssbeskrivelse endpoint, takes this shape.

    Parts of it are later assigned to attributes on class-instances.
    """

    Uttaksbeskrivelse_lagd: str
    base: str
    TabellId: str
    Huvudtabell: str
    DeltabellTitler: list[dict[str, str]]
    deltabller: list[DelTabellType]
    null_prikk_missing_kodeliste: NotRequired[list[SuppressionCodeListType]]
    kodelister: list[KodelisteType]
    IRkodelister: NotRequired[list[KodelisteType]]



from __future__ import annotations

import logging
import sys
from typing import Any

from colorama import Back
from colorama import Fore
from colorama import Style


class ColoredFormatter(logging.Formatter):
    """Colored log formatter."""

    def __init__(
        self,
        *args: Any,  # noqa: ANN401
        colors: dict[str, str] | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize the formatter with specified format strings."""
        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record: logging.LogRecord) -> str:
        """Format the specified record as text."""
        record.color = self.colors.get(record.levelname, "")
        record.reset = Style.RESET_ALL

        return super().format(record)


formatter = ColoredFormatter(
    "{color} {levelname:8} {reset}| {message}",
    style="{",
    colors={
        "DEBUG": Fore.CYAN,
        "INFO": Fore.GREEN,
        "WARNING": Fore.MAGENTA,
        "ERROR": Fore.RED,
        "CRITICAL": Fore.RED + Back.WHITE + Style.BRIGHT,
    },
)
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.handlers[:] = []
logger.addHandler(handler)
logger.setLevel(logging.INFO)



from __future__ import annotations

import copy
import json
import math
from decimal import ROUND_HALF_UP
from decimal import Decimal
from decimal import localcontext
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from statbank.api_types import DelTabellType
    from statbank.api_types import FilBeskrivelseType
    from statbank.api_types import KodelisteTypeParsed
    from statbank.api_types import KolonneInternasjonalRapporteringType
    from statbank.api_types import KolonneStatistikkvariabelType
    from statbank.api_types import KolonneVariabelType
    from statbank.api_types import SuppressionCodeListType
    from statbank.api_types import SuppressionDeltabellCodeListType

import pandas as pd
import requests as r

from statbank.auth import StatbankAuth
from statbank.statbank_logger import logger
from statbank.uttrekk_validations import StatbankUttrekkValidators
from statbank.uttrekk_validations import StatbankValidateError


class StatbankUttrekksBeskrivelse(StatbankAuth, StatbankUttrekkValidators):
    """Class for talking with the "uttrekksbeskrivelses-API", which describes metadata about shape of data to be transferred.

    And metadata about the table itself in Statbankens system,
    like ID, name of codelists etc.


    Attributes:
        url (str): Main url for transfer
        time_retrieved  (str): Time of getting the Uttrekksbeskrivelse
        tableid (str): Originally the ID of the main table, which to get the
            Uttrekksbeskrivelse on, but is reset based on the info in the Uttrekksbeskrivelse.
            To compansate for the possibility of the user sending in "tablename"-name as tableid.
        tablename (str): The name of the main table in Statbanken, not numbers, like the ID is.
        subtables (dict): Names and descriptions of the individual "table-parts"
            that needs to be sent in as different DataFrames.
        variables (dict): Metadata about the columns in the different table-parts.
        codelists (dict): Metadata about column-contents, like formatting on time, or possible values ("codes").
        suppression (dict): Details around extra columns which describe main column's "prikking", meaning their suppression-type.
        headers (dict): The headers for the request, might be sent in from a StatbankTransfer-object.
        filbeskrivelse (dict): The "raw" json returned from the API-get-request, loaded into a dict.

    """

    def __init__(
        self,
        tableid: str,
        raise_errors: bool = False,
        headers: dict[str, str] | None = None,
    ) -> None:
        """Makes a request to the Statbank-API, populates the objects attributes with parts of the return values."""
        self.url = self._build_urls()["uttak"]
        self.time_retrieved = ""
        self.tableid = tableid
        if not isinstance(raise_errors, bool):
            error_msg = (  # type: ignore[unreachable]
                "raise_errors must be a bool, the loaduser parameter has been removed."
            )
            raise TypeError(error_msg)
        self.raise_errors = raise_errors
        self.tablename = ""
        self.subtables: dict[str, str] = {}
        self.variables: list[DelTabellType] = []
        self.codelists: dict[str, KodelisteTypeParsed] = {}
        self.suppression: None | list[SuppressionCodeListType] = None
        if headers:
            self.headers = headers
        else:
            self.headers = self._build_headers()
        try:
            self._get_uttrekksbeskrivelse()
        finally:
            if hasattr(self, "headers"):
                del self.headers
        self._split_attributes()

    def __str__(self) -> str:
        """Returns a string representation of the object, which is the Uttrekksbeskrivelse."""
        variabel_text = ""
        for i, deltabell in enumerate(self.variables):
            variabel_text += f"""\nDeltabell (DataFrame) nummer {i+1}:
                {deltabell["deltabell"]}
                """
            variables: list[
                KolonneVariabelType
                | KolonneStatistikkvariabelType
                | KolonneInternasjonalRapporteringType
                | SuppressionDeltabellCodeListType
            ] = [*deltabell["variabler"], *deltabell["statistikkvariabler"]]
            if "null_prikk_missing" in deltabell:
                variables += deltabell["null_prikk_missing"]
            if "internasjonal_rapportering" in deltabell:
                variables += deltabell["internasjonal_rapportering"]

            variabel_text += f"Antall kolonner: {len(variables)}"
            for j, variabel in enumerate(variables):
                variabel_text += f"\n\tKolonne {j+1}: "
                variabel_text += str(variabel.get("Kodeliste_text", ""))
                variabel_text += str(variabel.get("Text", ""))
                supp = variabel.get("gjelder_for_text", "")
                if supp:
                    variabel_text += f"Suppressionfo column {variabel.get('gjelder_for__kolonner_nummer')}: {supp}"
            variabel_text += f'\nEksempellinje: {deltabell["eksempel_linje"]}'

        mult_codelists = math.prod([len(x["koder"]) for x in self.codelists.values()])
        variabel_text += f'\n"Ekspandert matrise/antall koder i kodelistene ganget med hverandre er: {mult_codelists}'

        return f"""Uttrekksbeskrivelse for statbanktabell {self.tableid}.

        Hele filbeskrivelsen "rå" ligger under .filbeskrivelse
        Andre attributter:
        .subtables, .codelists, .suppression, .variables
{variabel_text}
        """

    def __repr__(self) -> str:
        """Return a string representation of how to instantiate this object again."""
        return f'StatbankUttrekksBeskrivelse(tableid="{self.tableid}",)'

    def transferdata_template(
        self,
        dfs: list[pd.DataFrame] | None = None,
    ) -> dict[str, str] | dict[str, pd.DataFrame]:
        """Get the shape the data should have to name the "deltabeller".

        If we didnt use a dictionary we would have to rely on the order of a list of "deltabeller".
        Instead we chose to explicitly name the deltabller in this package, and make you check this after creation.

        Parameters
        -------
        dfs: if provided, will try to use pandas dataframes sent in to populate the dict returned by the method.
            Send in one dataframe, several, a list of dataframes or similar.
            ORDER IS IMPORTANT make sure the result is what you expect.

        Returns:
        -------
        A dictionary with correct keys, but placeholders for where the dataframes should go if no Dataframes are passed.
        A dict of dataframes as values if a list of Dataframes are sent in, or dataframes as individual parameters.
        """
        # If sending in a list, unwrap one layer
        if dfs is not None:
            if isinstance(dfs, pd.DataFrame):  # type: ignore[unreachable]
                dfs = [dfs]  # type: ignore[unreachable]
            if not all(isinstance(df, pd.DataFrame) for df in dfs):
                error_msg = "All elements sent in to transferdata_template must be pandas dataframes."
                raise TypeError(error_msg)
            if len(dfs) != len(self.subtables):
                error_msg = (
                    "Number of dataframes in must match the number of subtables."
                )
                raise KeyError(error_msg)
            template = {k: dfs[i] for i, k in enumerate(self.subtables.keys())}
            msg = "{\n"
            for k, v in template.items():
                msg += f'"{k}" : Dataframe with column-names: {v.columns}\n'
            msg += "}"
            logger.info(msg)
            return template
        non_df_template = {k: f"df{i}" for i, k in enumerate(self.subtables.keys())}
        logger.info(
            """Your template should look like this: %s
            You can also send in a list of dataframes to this function, and get a dict back, but check the order!""",
            non_df_template,
        )
        return non_df_template

    def to_json(self, path: str = "") -> None | str:
        """Store a copy of the current state of the uttrekk-object as a json.

        If path is provided, tries to write to it,
        otherwise will return a json-string for you to handle like you wish.

        Args:
            path (str): if provided, will try to write a json to a local path

        Returns:
            None | str: If path is provided, tries to write a json to a file and returns nothing.
                If path is not provided, returns the json-string for you to handle as you wish.
        """
        # Need to this because im stupidly adding methods from other class as attributes
        content = {k: v for k, v in self.__dict__.items() if not callable(v)}

        if path:
            logger.info("Writing to %s", path)
            with Path(path).open(mode="w") as json_file:
                json_file.write(json.dumps(content))
        else:
            return json.dumps(content)
        return None

    def validate(
        self,
        data: dict[str, pd.DataFrame],
        raise_errors: bool = False,
    ) -> dict[str, ValueError]:
        """Uses the contents of itself to validate the data against.

        All validation happens locally, so dont be afraid of any data
        being sent to statbanken using this method.

        Args:
            data (dict[str, pd.DataFrame]): The data to validate in a dictionary of deltabell-names as keys and pandas-dataframes as values.
            raise_errors (bool): True/False based on if you want the method to raise its own errors or not.

        Returns:
            dict[str, ValueError]: A dictionary of the errors the validation wants to raise.

        Raises:
            StatbankValidateError: if raise_errors is set to True and there are validation errors.
        """
        if not raise_errors:
            raise_errors = self.raise_errors

        validation_errors: dict[str, ValueError] = {}
        logger.info("validating...")

        self._validate_number_dataframes(data=data)
        validation_errors = self._validate_number_columns(
            data,
            validation_errors,
        )
        validation_errors = self._category_columns_are_strings(data, validation_errors)
        validation_errors = self._category_code_usage(data, validation_errors)
        validation_errors = self._check_for_floats(data, validation_errors)
        validation_errors = self._check_for_literal_nans_in_strings(
            data,
            validation_errors,
        )
        validation_errors = self._check_statistikkvar_numerical(data, validation_errors)
        validation_errors = self._check_rounding(data, validation_errors)
        validation_errors = self._check_time_same_values_in_deltabeller(
            data,
            validation_errors,
        )
        validation_errors = self._check_time_formats(data, validation_errors)
        validation_errors = self._check_suppression(data, validation_errors)
        validation_errors = self._check_unique_combinations_categories_times(
            data,
            validation_errors,
        )

        if raise_errors and validation_errors:
            raise StatbankValidateError(list(validation_errors.values()))
        logger.info(
            "validation finished (if nothing is logged over info level, everything should be fine.)",
        )
        return validation_errors

    def get_totalcodes_dict(self) -> dict[str, str]:
        """Makes a dict from each codelist where a code for "totals" is included.

        Keys being the name of the codelist, values being the code to put into categorical columns, that describes totals.
        This dict can be passed into the parameters "fillna_dict" and "grand_total" in the function "agg_all_combos" in the package ssb-fagfunksjoner.

        Returns:
            dict[str, str]: A dictionary with the codelist-names as keys, the total-codes as values.
        """
        result: dict[str, str] = {}
        for name, kodeliste in self.codelists.items():
            if "SumIALtTotalKode" in kodeliste:
                result[name] = kodeliste["SumIALtTotalKode"]
        return result

    def round_data(
        self,
        data: dict[str, pd.DataFrame],
        round_up: bool = True,
    ) -> dict[str, pd.DataFrame]:
        """Converts all decimal numbers to strings, with the correct number of decimals.

        IMPORTANT: Rounds "real halves" (0.5) UP, instead of "to even numbers" like Python does by default.
        This is maybe the behaviour staticians are used to from Excel, SAS etc.

        Args:
            data (dict[str, pd.DataFrame]): The data to validate in a dictionary of deltabell-names as keys and pandas-dataframes as values.
            round_up (bool): Default behaviour is rounding up like Excel or SAS. Setting this to False will instead use Python's default "Round towards equal" / "Banker's rounding"

        Returns:
            dict[str, pd.DataFrame]: A dictionary in the same shape as sent in, but with dataframes altered to correct for rounding.
        """
        data_copy = copy.deepcopy(data)
        for deltabell in self.variables:
            deltabell_name = deltabell["deltabell"]
            for variabel in deltabell["statistikkvariabler"]:
                if "Antall_lagrede_desimaler" in variabel:
                    col_num = int(variabel["kolonnenummer"]) - 1
                    decimal_num = int(variabel["Antall_lagrede_desimaler"])
                    # Nan-handling?
                    if (
                        "float"
                        in str(data_copy[deltabell_name].dtypes.iloc[col_num]).lower()
                    ):  # If column is passed in as a float, we can handle it
                        logger.info(
                            "Rounding column %s in %s into a string, with %s decimals.",
                            col_num + 1,
                            deltabell_name,
                            decimal_num,
                        )
                        data_copy[deltabell_name][
                            data_copy[deltabell_name].columns[col_num]
                        ] = (
                            data_copy[deltabell_name]
                            .iloc[:, col_num]
                            .astype("Float64")
                            .apply(self._round, decimals=decimal_num, round_up=round_up)
                            .astype(str)
                            .str.replace("<NA>", "", regex=False)
                            .str.replace(".", ",", regex=False)
                        )
                    else:
                        logger.info(
                            "not a float %s: %s",
                            col_num,
                            str(data_copy[deltabell_name].dtypes.iloc[col_num]),
                        )
        return data_copy

    @staticmethod
    def _round(n: float, decimals: int = 0, round_up: bool = True) -> str:
        if pd.isna(n):
            result: str = ""
        elif round_up and decimals and (n or n == 0):
            with localcontext() as ctx:
                ctx.rounding = ROUND_HALF_UP
                result = str(round(Decimal(n), decimals))
        elif not round_up and decimals and (n or n == 0):
            result = str(round(Decimal(n), decimals))
        elif n or n == 0:
            result = str(Decimal(n).to_integral_value())
        else:
            result = str(n)
        return result

    def _get_uttrekksbeskrivelse(self) -> None:
        filbeskrivelse_url = self.url + "tableId=" + self.tableid
        try:
            filbeskrivelse_response = self._make_request(filbeskrivelse_url)
        finally:
            if hasattr(self, "headers"):
                del self.headers

        # Rakel encountered an error with a tab-character in the json, should we just strip this?
        filbeskrivelse_json = filbeskrivelse_response.text.replace("\t", "")
        # Also deletes / overwrites returned Auth-header from get-request
        filbeskrivelse: FilBeskrivelseType = json.loads(filbeskrivelse_json)
        logger.info(
            "Hentet uttaksbeskrivelsen for %s, med tableid: %s den %s",
            filbeskrivelse["Huvudtabell"],
            self.tableid,
            str(filbeskrivelse["Uttaksbeskrivelse_lagd"]),
        )

        # reset tableid and hovedkode after content of request
        self.filbeskrivelse = filbeskrivelse

    def _make_request(self, url: str) -> r.Response:
        response = r.get(url, headers=self.headers, timeout=10)
        response.raise_for_status()
        return response

    def _split_attributes(self) -> None:
        # tableid might have been "hovedkode" up to this point, as both are valid in the URI
        self.time_retrieved = self.filbeskrivelse["Uttaksbeskrivelse_lagd"]
        self.tableid = self.filbeskrivelse["TabellId"]
        self.tablename = self.filbeskrivelse["Huvudtabell"]
        self.subtables = {
            x["Filnavn"]: x["Filtext"] for x in self.filbeskrivelse["DeltabellTitler"]
        }
        self.variables = self.filbeskrivelse["deltabller"]
        self.codelists = {}
        if "kodelister" in self.filbeskrivelse:
            kodelister = self.filbeskrivelse["kodelister"]
            if "IRkodelister" in self.filbeskrivelse:
                kodelister = [*kodelister, *self.filbeskrivelse["IRkodelister"]]
            for kodeliste in kodelister:
                new_kodeliste = {}
                for kode in kodeliste["koder"]:
                    new_kodeliste[kode["kode"]] = kode["text"]
                self.codelists[kodeliste["kodeliste"]] = {"koder": new_kodeliste}
                if "SumIALtTotalKode" in kodeliste:
                    self.codelists[kodeliste["kodeliste"]]["SumIALtTotalKode"] = (
                        kodeliste["SumIALtTotalKode"]
                    )

        if "null_prikk_missing_kodeliste" in self.filbeskrivelse:
            self.suppression = self.filbeskrivelse["null_prikk_missing_kodeliste"]
